export class Type {
   constructor(f, options = {}) {
      this.test = f
      this.from = options.from ?? {patterns: []} // mock empty Implementations
   }
   toString() {
      return this.name || `[Type ${this.test}]`
   }
}

export const Undefined = new Type(t => typeof t === 'undefined')
export const TypeOfTypes = new Type(t => t instanceof Type)

export const Returns = (type, f) => (f.returns = type, f)

export const whichType = typs => Returns(TypeOfTypes, item => {
   for (const type of Object.values(typs)) {
      if (!(type instanceof Type)) continue
      if (type.test(item)) return type
   }
   let errorMsg = ''
   try {
      errorMsg = `no known type for '${item}'` // fails for Symbols, e.g.
   } catch {
      errorMsg = `no known type for '${item.toString()}'`
   }
   throw new TypeError(errorMsg)
})

export const typeOf = math => whichType(math.types)

// bootstrapping order matters, but order of exports in a module isn't
// simply the order that the items are listed in the module. So we make
// an explicitly ordered export of implementations for this sake:

export const bootstrapTypes = {
   Type, Undefined, TypeOfTypes, typeOf
}
