import assert from 'assert'
import math from '#nanomath'
import {NumberT} from '#number/NumberT.js'

import {Returns} from '../Type.js'
import {isPlainFunction} from '../helpers.js'

describe('Core types', () => {
   it('creates an object with all of the types', () => {
      assert('NumberT' in math.types)
      assert.strictEqual(math.types.NumberT, NumberT)
      assert('Undefined' in math.types)
      assert(!('Type' in math.types))
      assert(math.types.Undefined.test(undefined))
      assert(!math.types.Undefined.test(3))
      assert(math.types.TypeOfTypes.test(math.types.Undefined))
   })

   it('supports a typeOf operator', () => {
      const tO = math.typeOf
      assert.strictEqual(tO(7), NumberT)
      assert.strictEqual(tO(undefined), math.types.Undefined)
      assert.strictEqual(tO(NumberT), math.types.TypeOfTypes)
      assert.throws(() => tO(Symbol()), TypeError)
   })

   it('lets you create a new Type', () => {
      const Foo = new math.Type(
         item => item && typeof item === 'object' && 'bar' in item)
      assert(Foo.test({bar: 'baz'}))
      math.merge({Foo})
      assert.strictEqual(math.types.Foo, Foo)
      assert.strictEqual(math.typeOf({gold: 7, bar: 3}), Foo)
   })

   it('provides a return-value labeling', () => {
      const labeledF = Returns(math.types.Undefined, () => undefined)
      assert.strictEqual(typeof labeledF, 'function')
      assert.strictEqual(labeledF.returns, math.types.Undefined)
      assert(isPlainFunction(labeledF))
   })

})
