import assert from 'assert'
import {TypeDispatcher} from '../TypeDispatcher.js'
import * as booleans from '#boolean/all.js'
import * as generics from '#generic/all.js'
import * as numbers from '#number/all.js'
import {onType} from "#core/helpers.js"
import {Any} from "#core/TypePatterns.js"
import {Returns} from "#core/Type.js"
import {plain} from "#number/helpers.js"

describe('TypeDispatcher', () => {
   it('can build in stages', () => {
      const incremental = new TypeDispatcher()
      assert.strictEqual(
         incremental.typeOf(undefined), incremental.types.Undefined)
      incremental.merge(numbers)
      const {NumberT, TypeOfTypes, Undefined} = incremental.types
      assert(NumberT.test(7))
      assert.strictEqual(incremental.add(-1.5, 0.5), -1)
      // Make Undefined act like zero:
      incremental.merge({add: onType(
         [Undefined, Any], (_m, [_U, T]) => Returns(T, (_a, b) => b),
         [Any, Undefined], (_m, [T]) => Returns(T, a => a)
      )})
      assert.strictEqual(incremental.add(7, undefined), 7)
      assert.strictEqual(
         incremental.resolve('add', [TypeOfTypes, Undefined]).returns,
         TypeOfTypes)
      assert.strictEqual(incremental.add(undefined, -3.25), -3.25)
      assert.strictEqual(
         incremental.add.resolve([Undefined, NumberT]).returns,
         NumberT)
      // Oops, changed my mind, make it work like NaN with numbers:
      const alwaysNaN = Returns(NumberT, () => NaN)
      incremental.merge({add: onType(
         [Undefined, NumberT], alwaysNaN,
         [NumberT, Undefined], alwaysNaN
      )})
      assert(isNaN(incremental.add(undefined, -3.25)))
      assert.strictEqual(
         incremental.add.resolve([Undefined, NumberT]).returns,
         NumberT)
      assert.strictEqual(incremental.isnan(NaN), 1)
      incremental.merge(booleans)
      assert.strictEqual(incremental.boolean(undefined), false)
      assert.strictEqual(incremental.isnan(NaN), true)
   })
   it('changes methods when their dependencies change', () => {
      const gnmath = new TypeDispatcher(generics, numbers)
      assert.strictEqual(gnmath.square(-2.5), 6.25)
      // change multiplication of number to floor its first argument (?!)
      gnmath.merge({multiply: plain((a,b) => Math.floor(a) * b)})
      assert.strictEqual(gnmath.square(-2.5), 7.5)
   })
   it('detects dependencies on conversion operations', () => {
      const bgn = new TypeDispatcher(booleans, generics, numbers)
      const {BooleanT, NumberT} = bgn.types
      assert(!bgn._behaviors.negate.has([BooleanT]))
      assert.strictEqual(bgn.negate(true), -1)
      assert(bgn._behaviors.negate.has([BooleanT]))
      const deps = bgn._dependencies.negate
      bgn.merge({number: onType([BooleanT], Returns(NumberT, b => b ? 2 : 0))})
      assert(!bgn._behaviors.negate.has([BooleanT]))
      assert.strictEqual(bgn.negate(true), -2)
   })
})
