import assert from 'assert'
import {
   pattern, Any, Multiple, Optional, needsCollection
} from '../TypePatterns.js'
import {Undefined, TypeOfTypes} from '../Type.js'

describe('Type patterns', () => {
   it('makes patterns from types and sequences thereof', () => {
      const undef1 = pattern(Undefined)
      assert.deepStrictEqual(undef1.match([Undefined]), [1, Undefined])
      assert.deepStrictEqual(
         undef1.match([Undefined, Undefined]), [1, Undefined])
      assert.deepStrictEqual(undef1.match([TypeOfTypes]), [-1, Undefined])
      assert.deepStrictEqual(undef1.match([]), [-1, Undefined])
      assert.deepStrictEqual(undef1.sampleTypes(), [Undefined])
      assert(undef1.equal(pattern(Undefined)))
      assert(!undef1.equal(pattern(TypeOfTypes)))
      const tu2 = pattern([TypeOfTypes, undef1])
      assert.deepStrictEqual(
         tu2.match([Undefined, TypeOfTypes]), [-1, Undefined])
      assert.deepStrictEqual(
         tu2.match([TypeOfTypes, Undefined]), [2, [TypeOfTypes, Undefined]])
   })

   it('allows Any as a wildcard', () => {
      const midWild = pattern([Undefined, Any, TypeOfTypes])
      assert.deepStrictEqual(
         midWild.match([Undefined, Undefined, TypeOfTypes]),
         [3, [Undefined, Undefined, TypeOfTypes]])
      assert.deepStrictEqual(
         midWild.match([Undefined, TypeOfTypes, TypeOfTypes]),
         [3, [Undefined, TypeOfTypes, TypeOfTypes]])
      assert.deepStrictEqual(
         midWild.match([Undefined, TypeOfTypes, Undefined]),
         [-1, Undefined])
      assert.deepStrictEqual(
         midWild.match([Undefined, TypeOfTypes]),
         [-1, Undefined])
      assert.deepStrictEqual(
         midWild.sampleTypes(), [Undefined, Undefined, TypeOfTypes])
      assert(midWild.equal(midWild))
      assert(!midWild.equal(pattern([Any, Any, TypeOfTypes])))
   })

   it("supports greedy multiple and optional parameters", () => {
      const midOpt = pattern([Undefined, Optional(TypeOfTypes), Undefined])
      assert.deepStrictEqual(
         midOpt.match([Undefined, Undefined]), [2, [Undefined, [], Undefined]])
      assert.deepStrictEqual(
         midOpt.match([Undefined, TypeOfTypes, Undefined]),
         [3, [Undefined, [TypeOfTypes], Undefined]])
      assert.deepStrictEqual(
         midOpt.match([Undefined, TypeOfTypes, TypeOfTypes]),
         [-1, Undefined])
      assert.deepStrictEqual(midOpt.sampleTypes(), [Undefined, Undefined])
      const midMulti = pattern([Undefined, Multiple(TypeOfTypes), Undefined])
      assert.deepStrictEqual(
         midMulti.match([Undefined, Undefined]),
         [2, [Undefined, [], Undefined]])
      assert.deepStrictEqual(
         midMulti.match([Undefined, TypeOfTypes, Undefined]),
         [3, [Undefined, [TypeOfTypes], Undefined]])
      assert.deepStrictEqual(
         midMulti.match(
            [Undefined, TypeOfTypes, TypeOfTypes, TypeOfTypes, Undefined]),
         [5, [Undefined, [TypeOfTypes, TypeOfTypes, TypeOfTypes], Undefined]])
      assert.deepStrictEqual(
         midMulti.match([Undefined, TypeOfTypes, TypeOfTypes]),
         [-1, Undefined])
      assert.deepStrictEqual(midOpt.sampleTypes(), [Undefined, Undefined])
      assert(!midMulti.equal(midOpt))
      const whyNot = pattern(
         [Undefined, Multiple([Undefined, TypeOfTypes]), TypeOfTypes])
      assert.deepStrictEqual(
         whyNot.match([Undefined, TypeOfTypes]),
         [2, [Undefined, [], TypeOfTypes]])
      assert.deepStrictEqual(
         whyNot.match([Undefined, Undefined, TypeOfTypes, TypeOfTypes]),
         [4, [Undefined, [[Undefined, TypeOfTypes]], TypeOfTypes]])
      assert.deepStrictEqual(
         whyNot.match([
            Undefined,
            Undefined, TypeOfTypes, Undefined, TypeOfTypes,
            TypeOfTypes
         ]),
         [6, [
            Undefined,
            [[Undefined, TypeOfTypes], [Undefined, TypeOfTypes]],
            TypeOfTypes
         ]])
      assert.deepStrictEqual(
         whyNot.match([Undefined, TypeOfTypes, Undefined, TypeOfTypes]),
         [2, [Undefined, [], TypeOfTypes]])
      assert.deepStrictEqual(
         whyNot.sampleTypes(), [Undefined, TypeOfTypes])
      assert(whyNot.equal(whyNot))
   })
   it('determines whether a template needs a collection function', () => {
      assert(!needsCollection([Undefined]))
      assert(needsCollection([Undefined, [Undefined, Undefined]]))
      assert(needsCollection(
         [Undefined, {actual: Undefined, matched: TypeOfTypes}]))
   })
})
