import assert from 'assert'
import {
   Implementations, onType, isPlainObject, isPlainFunction
} from '../helpers.js'
import {Type, Undefined, TypeOfTypes} from '../Type.js'
import {TypePattern} from '../TypePatterns.js'

describe('Core helpers', () => {
   it('defines what Implementations are', () => {
      const imps = onType(Undefined, 7, [TypeOfTypes, Undefined], -3)
      assert(imps instanceof Implementations)
      assert(imps.patterns instanceof Array)
      assert(imps.patterns.every(([k]) => k instanceof TypePattern))
   })

   it('detects plain objects', () => {
      assert(!isPlainObject(Undefined))
      assert(!isPlainObject([2, 3, 4]))
      assert(!isPlainObject('pat'))
      assert(isPlainObject({plain: 'object'}))
      assert(isPlainObject({}))
   })

   it('detects plain functions (not constructors)', () => {
      function embedded() {return 1}
      class Embedded {}
      assert(isPlainFunction(embedded))
      assert(isPlainFunction(() => 'wow'))
      assert(isPlainFunction(it))
      assert(!isPlainFunction(7))
      assert(!isPlainFunction(Implementations))
      assert(!isPlainFunction(Type))
      assert(!isPlainFunction(Embedded))
   })
})
