import {pattern} from './TypePatterns.js'

export class Implementations {
   constructor(imps) {
      this.patterns = []
      this._add(imps)
   }
   _add(imps) {
      for (let i = 0; i < imps.length; ++i) {
         this.patterns.push([pattern(imps[i]), imps[++i]])
      }
   }
   also(...imps) {
      this._add(imps)
   }
}

export const onType = (...imps) => new Implementations(imps)

export const isPlainObject = obj => {
   if (typeof obj !== 'object') return false
   if (!obj) return false // excludes null
   const proto = Object.getPrototypeOf(obj)
   return !proto || proto === Object.prototype
}

// Class entities have typeof === 'function', but mostly we do not
// want to treat them like ordinary functions, hence we need some way
// of telling. There doesn't seem to be a definitive way, so we
// use a mishmash of techniques from
// https://stackoverflow.com/questions/526559
const isClass = f => {
   if (!f.prototype) return false
   if (f.constructor !== Function) return false
   const protoCtor = f.prototype.constructor
   if (protoCtor
       && protoCtor.toString
       && protoCtor.toString().startsWith('class')
   ) {
      return true
   }
   if (Object.getPrototypeOf(f) !== Function.prototype) return true
   if (Object.getOwnPropertyNames(f).includes('arguments')) return false
   return Object.getOwnPropertyNames(f.prototype).length > 1
}

// returns true for functions that are not class constructors, at
// least so far as we can tell:
export const isPlainFunction = f => {
   return typeof f === 'function' && !isClass(f)
}
